#include <mrt.h>
#include "propertyList.h"
#include "salamanderInterface.h"

#define DEFAULT_SALY_PORT     9000
#define DEFAULT_SALY_DATASIZE 2048
#define DEFAULT_SALY_NUM_MSG  10
#define DEFAULT_TRACE_STR     "crush.log"

// getopt stuff
extern char	*optarg;
extern int	optind,
		opterr,
		optopt;

char		USAGE_STRING[] =
			"\n%s is used to exercise the salamander server\n"
			"\nUsage:\n%s [-s <server>] [-p <port>] "
			"[-f <filename>] [-d <data size>]\n"
			"\t\t[-n <# of msgs] [-l <logfile>] "
			"[-k] <channel>\n\n"
			"\t-s specify salamander server\n"
			"\t-p specify port number"
			"\t(Default %u)\n"
			"\t-f specify filename\n"
			"\t-d specify data size"
			"\t(Default: %u)\n"
			"\t-n specify number of messages"
			"\t(Default: %u)\n"
			"\t-l specify logfile"
			"\t(Default: %s)\n"
			"\t-k persistent\n\n";

void	PrintUsage(const char *ProgramNamePtr);

int
main (int argc, char *argv[]) {
	salamander_t	*sal;
	trace_t		*TracePtr;
	char			*SalyServerStr;
	char			*SalyFileStr;
	char			*ChannelStr;
	char			*TraceStr;
	u_short			SalyPort;
	prefix_t		*server_prefix;
	plist_t		plist;
	plist_t		rplist;
	struct fd_set		RFDSet;
	struct fd_set		WFDSet;
	struct timeval		TimeVal;
	void			*DataPtr;
	void			*rDataPtr;
	u_long		DataLen;
	u_long		rDataLen;
	u_short		Errors;
	int			KeepOpen = 0;
	int			i;
	int			fd;
	int			Max;
	int			SentRecv;
	int			ret;
	int			UseTCP;
	int			UseFile;

	char			c;

	SalyServerStr = NULL;
	SalyPort      = DEFAULT_SALY_PORT;
	SalyFileStr   = NULL;
	DataLen       = DEFAULT_SALY_DATASIZE;
	Max           = DEFAULT_SALY_NUM_MSG;
	TraceStr      = DEFAULT_TRACE_STR;
	ChannelStr    = NULL;
	server_prefix = NULL;
	UseFile       = 0;
	UseTCP        = 0;

	if (argc < 2) {
		PrintUsage(argv[0]);
		return 0xbadbad;
	}

	Errors = 0;
	while ((c = getopt(argc, argv, "s:p:f:d:n:l:k")) != -1) {
		switch (c) {
		case 's':
			SalyServerStr = optarg;
			if (UseFile)
				Errors++;
			UseTCP = 1;
			break;
		case 'p':
			SalyPort = atoi(optarg);
			if (UseFile)
				Errors++;
			UseTCP = 1;
			break;
		case 'f':
			SalyFileStr = optarg;
			if (UseTCP)
				Errors++;
			UseFile     = 1;
			break;
		case 'd':
			DataLen = atoi(optarg);
			break;
		case 'n':
			Max = atoi(optarg);
			if (Max <= 0)
				Errors++;
			break;
		case 'l':
			TraceStr = optarg;
			break;
		case 'k':
			KeepOpen = 1;
			break;
		}
	}
	for ( ; optind < argc; optind++)
		ChannelStr = argv[optind];

	// Do more error checking
	if (Errors)
		;
	else if (UseTCP && !SalyServerStr)
		Errors++;
	else if (UseFile && !SalyFileStr)
		Errors++;
	else if (!UseTCP && !UseFile)
		Errors++;
	else if (!ChannelStr)
		Errors++;

	if (Errors) {
		PrintUsage(argv[0]);
		return 0xbadbad;
	}

	// Init defaults
	TracePtr = New_Trace();
	set_trace(TracePtr, TRACE_FLAGS, TR_ERROR | NORM,
		TRACE_LOGFILE, TraceStr, NULL);
	init_mrt(TracePtr);

	// Send request
	if (!DataLen) {
		DataPtr = NULL;
		DataLen = 0;
	}
	else {
		DataPtr = malloc(DataLen);
		if (!DataPtr) {
			trace(NORM, TracePtr,
				"Error allocating memory\n");
			return 0xdabba;
		}
		else {
			memset(DataPtr, 'x', DataLen);
		}
	}
	plist   = createPropertyList();
	if (!plist) {
		return 0xbad;
	}
	updateProperty(plist, "COMMAND", "PUBLISH");
	updateProperty(plist, "NAME", ChannelStr);
	updateProperty(plist, "ID", "1");
	dumpPropertyList(plist);
	TimeVal.tv_sec  = 0;
	TimeVal.tv_usec = 500;

	if (UseFile) {
		// Use File Descriptor
		fd = open(SalyFileStr,
			O_RDWR | O_CREAT | O_APPEND);
		if (fd == -1) {
			trace(NORM, TracePtr, "Error connecting to server\n");
			return 0xbadbad;
		}
		sal = New_SalamanderFile(fd);
	}
	else {
		// Use TCP
		server_prefix = string_toprefix(SalyServerStr, NULL);
		if (!server_prefix) {
			trace(NORM, TracePtr, "Error finding server\n");
			return 0xdeadbeef;
		}

		// Open connection
		sal = New_Salamander(SalyPort);
		ret = connectToSalamanderServer(sal, server_prefix);
		if (ret != SALAMANDER_OK) {
			trace(NORM, TracePtr, "Error connecting to server\n");
			return 0xbadbad;
		}
		fd = sal->fd;
	}

	i = 0;
	SentRecv = 0;
	do {
		FD_ZERO(&RFDSet);
		FD_ZERO(&WFDSet);
		FD_SET(fd, &RFDSet);
		FD_SET(fd, &WFDSet);
		select(1, &RFDSet, &WFDSet, NULL, &TimeVal);
			
//		if (FD_ISSET(fd, &WFDSet) && (SentRecv < 256)) {
			ret = salamanderSendServerData(
					sal, plist, DataPtr, DataLen);
			if (ret != SALAMANDER_OK) {
				trace(NORM, TracePtr, "Error sending data\n");
				return 0xabbadaba;
			}
			else {
				trace(NORM, TracePtr, "Done sending data\n");
				SentRecv++;
			}
//			dumpPropertyList(plist);
//		}
		if (UseTCP && FD_ISSET(fd, &RFDSet) && (SentRecv > 0)) {
			ret = salamanderReadServerData(
					sal, &rplist, &rDataPtr, &rDataLen);
			if (ret != SALAMANDER_OK) {
				trace(NORM, TracePtr, "Error reading data\n");
				return 0xabbadaba;
			}
			else {
				trace(NORM, TracePtr, "Done reading data\n");
//				dumpPropertyList(rplist);
				if (rDataLen)
					free(rDataPtr);
//				destroyPropertyList(rplist);
				SentRecv--;
			}
		}
	} while (KeepOpen || ++i < Max);
	destroyPropertyList(plist);

	disconnectFromSalamanderServer(sal);
	if (UseTCP)
		Delete_Prefix(server_prefix);

	return 0;
}

void
PrintUsage(const char *ProgramNamePtr) {
	printf(
		USAGE_STRING,
		ProgramNamePtr,
		ProgramNamePtr,
		DEFAULT_SALY_PORT,
		DEFAULT_SALY_DATASIZE,
		DEFAULT_SALY_NUM_MSG,
		DEFAULT_TRACE_STR);
}
