#define _TRACELOOP_C_

#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include "trace.h"
#include "hopList.h"
#include "traceloop.h"
#include "Globals.h"

#include <mrt.h>

#define	DEFAULT_DPORT 32768 + 666
#define MIN_SEND_PORT 16384
#define MAX_SEND_PORT 32768
#define GETHOST_BUFSIZE	512
// Max Timeouts = 3 * 3 (# of tries per TTL)
#define MAX_TIMEOUTS 9

int Timeout = 3;
int SendPort = MIN_SEND_PORT;

/***********************************************************************
Given a prefix, we will send out a traceroute with the given parameters
and return the results of our traceroute in the hoplistnode...
***********************************************************************/
HopListNode *
SendTraceRoute(prefix_t *host_prefix, int max_ttl, int datalen,
	int verbose) {

	HopListNode		*HopList;
	Proto			*pr = NULL;

	Proto proto_v4 =
	{icmpcode_v4, recv_v4, NULL, NULL, NULL, NULL, 0,
			IPPROTO_ICMP, IPPROTO_IP, IP_TTL};

#ifdef	IPV6
	Proto proto_v6 =
	{icmpcode_v6, recv_v6, NULL, NULL, NULL, NULL, 0,
			IPPROTO_ICMPV6, IPPROTO_IPV6, IPV6_UNICAST_HOPS};
#endif

	pr = &proto_v4;

	pr->sasend = prefix_tosockaddr(host_prefix);

	// Right now we are nixing the protocol independent stuff.
	// Apparently this is more complicated then I originally
	// thought.
	pr->salen  = sizeof(struct sockaddr_in);
	pr->sarecv = (struct sockaddr *)calloc(1, pr->salen);
	pr->salast = (struct sockaddr *)calloc(1, pr->salen);
	pr->sabind = (struct sockaddr *)calloc(1, pr->salen);

	HopList = traceloop(pr, max_ttl, datalen, verbose);

	if (pr->sabind)
		free(pr->sabind);
	if (pr->salast)
		free(pr->salast);
	if (pr->sarecv)
		free(pr->sarecv);
	if (pr->sasend)
		free(pr->sasend);

	return HopList;
}


HopListNode *
traceloop(Proto *pr, int max_ttl, int datalen, int verbose) {
	int		ttl,
			probe,
			Timeouts;
	int seq, code, done;
	double rtt;
	struct rec *rec;
	struct timeval tvrecv;

	char		*sendbuf,
			*recvbuf;

	HopListNode	HopListHead,
			*HopListPtr;
	int		ret;

	// Create a raw socket for receiving data
	ret = socket(pr->sasend->sa_family, SOCK_RAW, pr->icmpproto);
	if (ret < 0)
		return NULL;
	
	pr->recvfd = ret;

	// Create a socket for sending datagrams
	ret = socket(pr->sasend->sa_family, SOCK_DGRAM, 0);
	if (ret < 0) {
		close(pr->recvfd);
		return NULL;
	}
	pr->sendfd = ret;

	pr->sabind->sa_family = pr->sasend->sa_family;
	// our source UDP port#
	pr->sport = (getpid() & 0xffff) | 0x8000;
	// our destination UDP port#
	pr->dport = SendPort;
	SendPort += 256;
	if (SendPort > MAX_SEND_PORT)
		SendPort = MIN_SEND_PORT;

	// Bind our sending socket
	sock_set_port(pr->sabind, pr->salen, htons(INADDR_ANY));
	ret = bind(pr->sendfd, pr->sabind, pr->salen);
	if (ret < 0) {
		close(pr->recvfd);
		close(pr->sendfd);
		return NULL;
	}
	ret = getsockname(pr->sendfd, pr->sabind, &pr->salen);
	if (ret < 0) {
		close(pr->recvfd);
		close(pr->sendfd);
		return NULL;
	}
	// Save the port number
	pr->sport = ntohs(((struct sockaddr_in *)pr->sabind)->sin_port);

	// Allocate some buffers
	sendbuf = (char *)calloc(1, BUFSIZE + datalen);
	recvbuf = (char *)calloc(1, BUFSIZE + datalen);

	seq = 0;
	done = 0;
	Timeouts = 0;

	// Initialize the list to contain the destination
	// of our traceroute.
	HopListHead.Next = (HopListNode *)calloc(1, sizeof(HopListNode));
	HopListPtr = HopListHead.Next;
	memcpy(&HopListPtr->Addr, pr->sasend, sizeof(struct sockaddr));
	
	for (ttl = 1; ttl <= max_ttl && done == 0; ttl++) {
		// Set the TTL level
		setsockopt(pr->sendfd, pr->ttllevel, pr->ttloptname,
			(char *)&ttl, sizeof(int));

		bzero(pr->salast, pr->salen);
//		printf("%p: %2d  ", &pr, ttl);
//		fflush(stdout);

		for (probe = 0; probe < NUM_PROBES; probe++) {
			rec = (struct rec *) sendbuf;
			rec->rec_seq = ++seq;
			rec->rec_ttl = ttl;
			gettimeofday(&rec->rec_tv, NULL);

			// Vary the port number of our destination
			sock_set_port(pr->sasend, pr->salen,
				htons(pr->dport + seq));
			sendto(pr->sendfd, sendbuf, datalen, 0,
				pr->sasend, pr->salen);

			HopListPtr->Next = (HopListNode *)
				calloc(1, sizeof(HopListNode));
			HopListPtr = HopListPtr->Next;

			code = (*pr->recv) (pr, seq, &tvrecv, recvbuf,
				datalen, verbose);
			if (code == -3) {
				/* timeout, no reply */
//				fprintf(stdout, " *");
				HopListPtr->RoundTripTime = 0;
				Timeouts++;
			}
			else {
				Timeouts = 0;
				memcpy(&HopListPtr->Addr, pr->sarecv,
					sizeof(struct sockaddr));

				if (sock_cmp_addr(pr->sarecv, pr->salast,
					pr->salen) != 0) {
					memcpy(pr->salast, pr->sarecv,
						pr->salen);
				}
				tv_sub(&tvrecv, &rec->rec_tv);
				rtt = tvrecv.tv_sec * 1000.0 +
					tvrecv.tv_usec / 1000.0;
				HopListPtr->RoundTripTime = rtt;
//				fprintf(stdout, "  %.3f ms", rtt);
//				fflush(stdout);
				// port unreachable; at destination
				if (code == -1)
					done++;
				else if (code >= 0) {
					;
//					fprintf(stdout, "%d: (ICMP %s)",
//						code, (*pr->icmpcode)(code));
				}
			}
//			fflush(stdout);
		}
		fprintf(stdout, "\n");
//		fflush(stdout);
		// We haven't received a response in the last 9
		// tries, so lets give up.
		if (Timeouts >= MAX_TIMEOUTS)
			break;
	}

	close(pr->sendfd);
	close(pr->recvfd);

//	PrintHopList(HopListHead.Next, pr->salen);
	free(sendbuf);
	free(recvbuf);

	return HopListHead.Next;
}

void
sock_set_port(struct sockaddr *sa, int salen, int port)
{
        switch (sa->sa_family) {
        case AF_INET: {
                struct sockaddr_in      *sin = (struct sockaddr_in *) sa;

                sin->sin_port = port;
                return;
        }

#ifdef  IPV6
        case AF_INET6: {
                struct sockaddr_in6     *sin6 = (struct sockaddr_in6 *)
sa;

                sin6->sin6_port = port;
                return;
        }
#endif
        }

    return;
}


int
sock_cmp_addr(const struct sockaddr *sa1,
	const struct sockaddr *sa2, int salen) {

	if (sa1->sa_family != sa2->sa_family)
		return(-1);

	return memcmp(
		&((struct sockaddr_in *) sa1)->sin_addr,
		&((struct sockaddr_in *)sa2)->sin_addr,
		sizeof(struct in_addr));
}

